/***************************************************************************************************/
/***************************************************************************************************/
//! \file a_star.h
//! \brief Dfini l'algorithm A* qui recherche le chemin le plus court ds un graph
//!    
//!	Auteur : Adrien Auclair
//!
/***************************************************************************************************/
/***************************************************************************************************/
#ifndef A_STAR_H
#define A_STAR_H


/*------------------------------------  Prototypes  ---------------------------------------------*/
class cA_Star_Node;
class cList_A_Star_Nodes;
class cA_Star_Priority_List;
class cA_Star_Search;

//#define DEBUG_A_STAR         // si dfini, l'algo envoie vers std::cout de infos sur sa progression
/*------------------------------------   Includes   ---------------------------------------------*/
#include <vector>
#include <assert.h>
#include <list>

#include "geo_utils.h"



/*-------------------------------------   Classes   ---------------------------------------------*/


/*************************************************************************************************/
//! definit une liste de pointeur sur des noeuds cA_Star_Node pour le A*
/*************************************************************************************************/
class cList_A_Star_Nodes : public std::list<cA_Star_Node*>
{
public : 
	/// retourne vrai si la liste contient le pointeur 'node' false sinon
	bool Contain( cA_Star_Node* node);	
};


/*************************************************************************************************/
//! un 'cA_Star_Node' est un noeud du graphe que parcoure l'algorithme A*
/*! classe abstraite
**************************************************************************************************/
class cA_Star_Node
{
public:	
	/// constructeur
	cA_Star_Node()                   
	{ 
		Path_Parent=NULL;
		Cost_From_Start=0;
		Estimated_Cost_To_Goal=0; 
	};

	cA_Star_Node * Path_Parent;      ///< le chemin vers le noeud par lequel on est arriv sur ce noeud
	float Cost_From_Start;           ///< le cout du chemin entre le depart et ce noeud
	float Estimated_Cost_To_Goal;    ///< le cout estim entre ce noeud et le noeud d'arrive
	cPoint2D Position;               ///< la position de ce noeud

	/// retoune le cout total estim, c'est  dire le cout exact depuis le depart + le cout estim jusqu'a l'arrive
	inline float Get_Cost() 
	{ 
		return Cost_From_Start + Estimated_Cost_To_Goal;
	}

	/// cout reel de ce noeud au noeud 'destination'
	virtual float Get_Cost( cA_Star_Node * destination)            =0;
	/// cout estim de ce noeud a 'destination'
	virtual float Get_Estimated_Cost( cA_Star_Node * destination)  =0;
	/// met ds voisinage la liste des noeuds voisins 
	virtual void  Get_List_Neighbors( cList_A_Star_Nodes & voisinage )    =0 ;	
};




/*************************************************************************************************/
//! est une liste de pointeur sur des 'cA_Star_Node', trie selon l'ordre croissant de cout total (reel+estim)
/*************************************************************************************************/
class cA_Star_Priority_List : public std::list<cA_Star_Node*>
{
public:		
	/// ajoute le noeud 'news_node' dans la liste trie
	void Add(cA_Star_Node* new_node);  
	/// retourne vrai si 'node' appartient a la liste, false sinon
	bool Contain( cA_Star_Node* node); 
};



/*************************************************************************************************/
//! classe contenant l'algorithme A*
/*! cherche le plus court chemin dans un graphe
**************************************************************************************************/
class cA_Star_Search
{
public:
	/// met dans 'chemin' la liste des noeuds pour aller du noeud 'start' au noeud 'destination'
	/** le noeud 'start' n'est pas mis ds 'chemin' \n
	* le noeud 'goal' est mis ds chemin \n
	* retourne 'true' si un chemin a ete trouve, false sinon \n
	*/
	bool Find_Path(  cA_Star_Node* start , cA_Star_Node* goal , cList_A_Star_Nodes & chemin  );
private:
	/// met ds 'chemin' la liste des nodes pour aller  destination
	/** ralis car 'destination' a un champ 'Path_Parent' qui indique par quel noeud le chemin arrive a lui \n
	* le noeud initial n'est pas inclus ds le chemin !	
	*/
	void Construct_Path( cA_Star_Node* destination , cList_A_Star_Nodes & chemin );
};

#endif