/***************************************************************************************************/
/***************************************************************************************************/
//! \file bsp.h
//! \brief Dfini un algorithm generique pour crer une BSP
//!    
//!	Auteur : Adrien Auclair
//!
//! --> adapte pour diviser n'importe quel espace \n
//! necessite de fournir la classe qui coupe un sous-espace en deux sous-espaces \n
//! cette classe doit avoir les methodes suivantes: \n
//!                   divider->Get_Position_Divider_And_Split( A , B , C ) \n
//!               qui retourne la position du divider A par rapport a 'divider' \n
//!               et subdivise le divider A en deux divider B et C si besoin est \n
//!
//!   et aussi une fonction Get_Position_Point( item) qui retourne ou est l'item par rapport au divider \n
/***************************************************************************************************/
/***************************************************************************************************/
#ifndef BSP_TREE_H
#define BSP_TREE_H


/*-------------------------------------  Globals  -----------------------------------------------*/
typedef enum { DIVIDER_FRONT , DIVIDER_BACK , DIVIDER_CROSS , DIVIDER_ON} BSP_POSITION_DIVIDER;
typedef enum { POINT_ON , POINT_FRONT , POINT_BACK }BSP_POSITION_POINT;


/*------------------------------------  Prototypes  ---------------------------------------------*/
template< class Divider_Type>                                     class cBSP_Node;
template<class cItem ,  class cBSP_Node , class cDivider_Type >   class cBSP_Tree;

/*------------------------------------   Includes   ---------------------------------------------*/
#include <iostream>


#include "assert.h"

using namespace std;

/*-------------------------------------   Classes   ---------------------------------------------*/


/*************************************************************************************************/
//! un cBSP_Node est un noeud du bsp
/*! chaque cBSP_Node contient un divider qui divise l'espace en deux demi-espaces
**************************************************************************************************/
template< class Divider_Type>
class cBSP_Node 
{
public:	
	/// constructeur par defaut, cre un noeud non valide, Id=-1
	cBSP_Node() { Front=NULL;Back=NULL;Id=-1;};                      
	/// constructeur
	/** 'div' : le divider de ce noeud, le demi espace devant ce divider sera mis dans 'front' \n
	*                                                derriere ce divider sera mis dans back \n
	* 'id' : l'id a associer   ce noeud
	*/
	cBSP_Node(Divider_Type & div , int id) 
	{ 		
		Divider=div;	
		Front=NULL; 
		Back=NULL; 
		Id=id;
	};	
	
	/// destructeur
	~cBSP_Node() { delete Front; Front=NULL; delete Back; Back=NULL;};
	/// retourn l'id de ce noeud
	inline int Get_Id() { return Id;};                                
	/// fixe l'id de ce noeud  'id'	
	inline Set_Id(int id) { Id=id;};                                  
	/// affiche sur 'cout' les infos de ce noeud et ses fils
	// TODO : devrait etre virtuelle, mais si c le cas, ca plante a la compile
	virtual void Print()                                         
	{
		//cout << "[" << Id << "]:" << divider << "  fils:  [";
		if( Front != NULL )
			cout << Front->Get_Id();
		else
			cout << "empty";
		cout << "] <---> [";
		if( Back != NULL )
			cout << Back->Get_Id();
		else
			cout << "empty";
		cout << "] \n";

		if( Front != NULL)
			Front->Print();
		if( Back != NULL)
			Back->Print();
	}
	
	cBSP_Node * Front;               ///< le sous-arbre contenant le demi-espace devant ce divider
	cBSP_Node * Back;                ///<           "                  "         derriere    "
	Divider_Type  Divider;           ///< le divider de ce noeud	

protected:
	int Id;                          ///< id du noeud
};




/*************************************************************************************************/
//! un BSP tree permet de subdiviser un espace en sous-espaces
/*! 'cItem' : le type d'objet qui compose l'espace a diviser ( generique, pas forcement des points) \n
    'cBSP_Node' : le type de noeud dcrit plus haut \n
    'cDivider_Type' : le type de divider utilis pour diviser un espace en deux sous-espaces \n
     \n
   IMPORTANT : la class cDivider_Type fournie ici doit imprativement etre la meme que celle utilise pour \n
               instancier la classe cBSP_Node ! \n

**************************************************************************************************/
template<class cItem ,  class cBSP_Node , class cDivider_Type >
class cBSP_Tree
{
public:	
	/// constructeur
	cBSP_Tree()                                    
	{
		Root = NULL;
		N_Nodes=0;
	}
	
	/// destructeur
	~cBSP_Tree()                                   
	{ 
		if( Root != NULL) delete Root; 
		Root=NULL; 
	};

	/// retoune la racine de l'arbre 
	cBSP_Node * Get_Root() { return Root; };       
    
	/// Ajoute le divider  l'espace
	void Add_Divider( cDivider_Type & divider)     
	{	
		if( Root == NULL)
		{
			Root = new cBSP_Node( divider,N_Nodes);				
			N_Nodes++;
		}
		else
			Add_Divider( divider , Root );
	}

	/// imprime sur std::cout les infos sur ce BSP
	void Print()                                   
	{
		cout << "######### BSP TREE  ######### \n";
		cout << "N_Nodes = " << N_Nodes << "\n";
		if(Root != NULL)
			Root->Print();
		cout << "############################# \n";
	}	

	/// retourne le 'cBSP_Node' dans lequel se trouve le cItem 'item'
	/** si item est situ exactement  l'intrieur d'un sous-espaces, le noeud retourn sera une feuille \n
	* si item est situ exactement sur un divider, le noeud retourn ne sera pas une feuille !\n
	* la recherche se fait dans l'espace tout entier\n
	*/
	cBSP_Node * Where_Is( cItem & item ) 
	{  
		return Where_Is( item , Root ); 
	};

	/// retourne le 'cBSP_Node' dans lequel se trouve le cItem 'item', en forcant  retourner une feuille !
	/** si item est situ exactement  l'intrieur d'un sous-espaces, le noeud retourn sera une feuille \n
	* si item est situ exactement sur un divider, le noeud retourn sera la feuille juste devant le divider
	*/
	cBSP_Node * In_Which_Leaf_Is( cItem & item ) 
	{  
		return In_Which_Leaf_Is( item , Root ); 
	};

	/// applique la fonction callback  tous les noeuds ( et pas aux feuilles )
	void Apply_To_Each_Node( void (*Call_Back)(cBSP_Node * node) )
	{
		Recursive_Apply_To_Each_Node(Call_Back , Root );
	}

private:	
	cBSP_Node * Root;            ///< la racine de l'arbre
	int N_Nodes;                 ///< le nombre de noeud que compte l'arbre
	
	/// applique la fonction callback  tous les noeuds sous le noeud 'node' compris ( et pas aux feuilles )
	void Recursive_Apply_To_Each_Node(void (*Call_Back)(cBSP_Node * node), cBSP_Node * node )
	{
		if( node == NULL)
			return;
		
		//si ce n'est pas une feuille !
		if( node->Front != NULL && node->Back != NULL)
		{
			Call_Back(node);
			Recursive_Apply_To_Each_Node(Call_Back , node->Front );
			Recursive_Apply_To_Each_Node(Call_Back , node->Back );
		}		
	}

	//---------------------------------------------------------------------------------------------------
	//! identique a 'cBSP_Node * Where_Is( cItem & item )'
	//! sauf que la recherche se fait dans l'espace correspondant aux sous-espaces devant et derriere 'node'
	cBSP_Node * Where_Is( cItem & item , cBSP_Node * node )
	{
		assert( node != NULL);
		BSP_POSITION_POINT pos;

		// si node == NULL, cela signifie qu'on est arriv dans une feuille qui n'a pas ete dfini
		// ce BSP laisse l'implmentation des feuilles a l'utilisateur !
		if( node == NULL) return NULL;

		//si on est sur une feuille, on retourne la feuille
		if( node->Front == NULL &&  node->Back==NULL) return node;

		// sinon, on cherche la position de 'item' par rapport a ce divider, et on continue 
		// la recherche dans le bon sous-espace
		pos = node->Divider.Get_Position_Point( item );
		switch( pos)
		{
		case POINT_FRONT :
			return Where_Is(item ,  node->Front );
		case POINT_BACK :
			return Where_Is(item ,  node->Back );
		case POINT_ON :
			return node;
		default:
			cout << "ERROR 001 : cBSP_Node * Where_Is( cItem & item , cBSP_Node * node ) \n";
			break;
		}

		return NULL;  // pour eviter le warning 'path without return !"
	}

	//---------------------------------------------------------------------------------------------------
	//! identique a 'BSP_cBSP_Node * In_Which_Leaf_Is(cItem & item)'
	//! sauf que la recherche se fait dans l'espace correspondant aux sous-espaces devant et derriere 'node'
	cBSP_Node * In_Which_Leaf_Is( cItem & item , cBSP_Node * node )
	{
		BSP_POSITION_POINT pos;
		// si node == NULL, cela signifie qu'on est arriv sur une feuille qui n'a pas ete dfini
		// ce BSP laisse l'implmentation des feuilles a l'utilisateur !
		if( node == NULL)
			return NULL;

		//si on est sur une feuille, on retourne la feuille
		if( node->Front == NULL &&  node->Back==NULL)
			return node;

		// sinon, on cherche la position de 'item' par rapport a ce divider, et on continue 
		// la recherche dans le bon sous-espace
		pos = node->Divider.Get_Position_Point( item );
		switch( pos)
		{
		case POINT_FRONT :
		case POINT_ON :
			return In_Which_Leaf_Is(item ,  node->Front );
		case POINT_BACK :
			return In_Which_Leaf_Is(item ,  node->Back );		
		default:
			cout << "ERROR 001 : cBSP_Node * Where_Is( cItem & item , cBSP_Node * node ) \n";
			break;
		}		

		return NULL;  // pour eviter le warning 'path without return'
	}

	//---------------------------------------------------------------------------------------------------
	/// ajoute le divider 'divider' au sous-espaces sous le noeud 'node'
	void Add_Divider( cDivider_Type & divider , cBSP_Node * node )
	{
		BSP_POSITION_DIVIDER pos;
		cDivider_Type div_front;
		cDivider_Type div_back;
		assert( node != NULL);		
		
		pos = node->Divider.Get_Position_Divider_And_Split( divider , div_front , div_back);		
		
		switch( pos)
		{
		case DIVIDER_FRONT:				
			if( node->Front != NULL) 
			{
				Add_Divider( divider , node->Front );
			}
			else
			{					
				node->Front = new cBSP_Node( divider,N_Nodes);					
				N_Nodes++;
			}
			break;
		case DIVIDER_BACK:
			if( node->Back != NULL)
			{
				Add_Divider( divider , node->Back );
			}
			else
			{					
				node->Back = new cBSP_Node( divider,N_Nodes);					
				N_Nodes++;
			}
			break;
		case DIVIDER_CROSS:
			
			if( node->Front != NULL)
			{				
				Add_Divider( div_front , node->Front);
				
			}
			else
			{
				node->Front = new cBSP_Node( div_front,N_Nodes);					
				N_Nodes++;
			}

			if(node->Back != NULL)
			{				
				Add_Divider( div_back , node->Back);
			}
			else
			{
				node->Back = new cBSP_Node( div_back,N_Nodes);	
				N_Nodes++;
			}
			break;
		case DIVIDER_ON:
			cout << "ERROR : class cBSP_Tree : DIVIDER_ON is not managed in this bsp !!! \n";
			return;
			break;
		default:
			cout << "ERROR : class cBSP_Tree : unknown position of the divider \n";
			return;				
			break;
		}		
	}

	

	//---------------------------------------------------------------------------------------------------
	/// convertit un 'BSP_POSITION_DIVIDER' en chaine de caractere
	static char * Get_String_BSP_Position_Divider(BSP_POSITION_DIVIDER pos)
	{
		switch(pos)
		{
		case DIVIDER_FRONT :
			return "DIVIDER_FRONT";
		case DIVIDER_BACK : 
			return "DIVIDER_BACK";
		case DIVIDER_CROSS : 
			return "DIVIDER_CROSS";
		case DIVIDER_ON :
			return "DIVIDER_ON";
		}
		return "DIVIDER_unknown";
	}

	//---------------------------------------------------------------------------------------------------	
	/// convertit un 'BSP_POSITION_POINT' en chaine de caractere
	static char * Get_String_BSP_Position_Point(BSP_POSITION_POINT pos)
	{
		switch(pos)
		{
		case POINT_ON :
			return "POINT_ON";
		case POINT_FRONT :
			return "POINT_FRONT";
		case POINT_BACK :
			return "POINT_BACK";
		}
		return "POSITION_unknown";
	}
};




#endif