/***************************************************************************************************/
/***************************************************************************************************/
//! \file game_bsp.h
//! \brief Utilise la class generique de bsp de bsp.h pour notre jeu 2D
//!    
//!	Auteur : Adrien Auclair
//!
//! La classe cGame_World contiendra la description complete d'un niveau, bas sur un BSP \n
//! elle fournira toutes les fonctions pour questionner la geomtrie du jeu
/***************************************************************************************************/
/***************************************************************************************************/
#ifndef GAME_BSP_H
#define GAME_BSP_H

/*------------------------------------  Prototypes  ---------------------------------------------*/
class cGame_World;
class cSquare_Area;

/*------------------------------------   Includes   ---------------------------------------------*/
#include <vector>

#include "HV_bsp.h"
#include "game_graph.h"
#include "level.h"

/*************************************************************************************************/
//! decrit une region rectangulaire de l'espace non divise ds le BSP tree, cad une feuille
/*! herite de cHV_Node pour etre insr ds l'arbre \n
// convention : les murs et les portals d'un feuille appartiennent  la zone \n
//              ils sont situs  la limite du rectangle, mais  l'intrieur ! \n
//              cela signifie que l'intersection entre deux cSquare_Area est : \n
//                 -> soit vide \n
//                 -> soit compose des portals et des murs entre ces deux zones ! \n
**************************************************************************************************/
class cSquare_Area : public cHV_Node
{
	friend class cSquare_Area_Debug_Displayer;
public:	
	//! constructeur
	//! 'area' : cPositive_Rectangle qui represente la zone contenue ds cette feuille	
	cSquare_Area(cPositive_Rectangle& area):cHV_Node() 
	{ 
		Bounds = area; 
	};

	/// destructeur
	~cSquare_Area();             

	/// affiche sur std::cout les infos sur ce cSquare_Area
	void Print();
	
		
	/// divise ce cSquare_Area en deux cSquare_Area
	/** les deux objets cSquare_Area sont crs dans cette fonction \n
	* 'wall' : la division se fait selon la droite porteuse de ce mur \n
	* 'A1' : adresse du pointeur sur la premiere cSquare_Area cr ! \n
	* 'A2' : adresse du pointeur sur la premiere cSquare_Area cr ! \n
	*/
	void split( cHV_Wall & wall, cSquare_Area ** A1, cSquare_Area ** A2);		
	
	/// ajoute  cette piece un mur
	/** 'w' : segment qui represente le mur, doit etre contenue dans la frontiere interieure de cette piece 	*/
	void Add_Wall(cHV_Seg & w); 

	/// ajoute un portal  cette piece
	/** 'portal' : pointeur vers le portal qu'on rajoute */
	void Add_Portal(cPortal * portal);
	
	/// ajoute un portal paralel dans cette piece
	/** 'portal1' : pointeur vers le premier portal \n
	* 'portal2' : pointeur vers le second portal
	*/
	void Add_Paralells_Portals(cPortal * portal1 , cPortal * portal2);	
	
	/// cherche le centre de la zone
	/** 'p' : contiendra le centre de la zone */
	void Get_Centre(cPoint2D & p)
	{
		Bounds.Get_Centre(p);
		assert( Bounds.In(p) );
	};	

	/// cree un noeud entre deux portails paralelles d'un mur de cette piece
	/**
	* on cherche les portails paralelles de cette piece \n
	* lorsqu'on en trouve on cre un noeud entre eux
	*/
	void Manage_Paralell_Portals();

	cPositive_Rectangle Bounds;            ///< rectangle representant la zone
	std::vector<cHV_Seg*> List_Walls;      ///< la liste des segments qui forment les murs de la piece
	std::vector<cPortal*> List_Portals;    ///< la liste des portals de la piece
};




/*************************************************************************************************/
//! contient le BSP qui dcrit la zone de jeu et les fonctions membres sur la zone de jeu
/*************************************************************************************************/
class cGame_World
{	
	friend class cGame_World_Debug_Displayer;
public:	
	/// constructeur	
	cGame_World();                             
	/// destructeur
	~cGame_World();                            
	/// affiche ds std::cout la liste des feuilles et leurs infos
	void Print_Leaves();                       

	/// met ds Path le chemin de pA  pB
	/** retourne vrai si un chemin a ete trouv, false sinon*/
	bool Find_Path( cPoint2D& pA , cPoint2D& pB , std::vector<cPoint2D> & Path);
	
	/// lit le niveau du fichier donn, et cre le BSP dcor associ
	/** retourne false si le fichier est introuvable ou non valide */
	//bool Load_Level( char * filename);	
	
	/// retourne true si la boule de centre (x,y) de rayon r intersecte un mur du monde
	bool Do_Ball_Intersect_Walls( int x, int y, int r);

	/// retourne vrai si le segment intersecte au moins un mur
	/** s'il y a intersection, le point d'intersection est mis ds p_inter \n
	* ne test que la partie du segment a une distance au carr < square_test_range */
	bool Do_Segment_Intersect_Walls(cSeg& seg , cPoint2D& p_inter ,  int square_test_range);

	/// meme chose mais test le segment sur toute la longueur
	bool Do_Segment_Intersect_Walls(cSeg& seg , cPoint2D& p_inter  );

	/// decore l'arbre bsp et cre les liste de murs et de portails
	void Decorate();

	cHV_BSP_Tree My_BSP;                     ///< l'arbre BSP qui contient la monde		
	static cPositive_Rectangle Ecran;        ///< rectangle qui represente la zone de jeu entiere
	
private :	
	std::vector<cSquare_Area*> List_Areas;   ///< la liste des pieces du jeu
	std::vector<cPortal*> List_Portals;      ///< la liste des portals qui sont les noeuds pour le A* algorithm
	int N_Leaves;                            ///< nombres de pieces ds le jeu
	int N_Portals;                           ///< nombre de portals	
	bool bDecorated_With_Areas;              ///< indique que les feuilles de l'arbre ont ete insres	
	cA_Star_Search Path_Finder;              ///< class pour chercher un chemin dans un graphe ( A* )	
	
	/// parcours l'arbre et remplace chaque feuille par un cSquare_Area
	/** met dans 'List_Areas' la liste des areas du jeu */
	void Decorate_With_Areas();

	/// remplace les feuilles du sous-arbre 'node' par un cSquare_Area
	/** 'area' contient un pointeur vers un cSquare_Area qui represente ( node->front union node->back )	\n
	* fonction recursive */
	void Decorate_With_Areas( cHV_Node * node ,  cSquare_Area * area );

	/// parcours la liste des cSquare_Areas et ajoute a chacune ses murs et ses portals
	/** met dans 'List_Portals' la liste des portals du jeu \n
	* parcours toutes les pieces, et pour chaque cot de chaque piece, \n
	*          ajoute au jeu les murs et les portals de ce cot */
	void Decorate_With_Walls_And_Portals();		

	/// declare au jeu l'existence d'un portal ( qui a peut-etre dja t cr )
	/** 'seg'  : segment porteur du nouveau portal \n
	*   'area' : une des deux cSquare_Area qui bordent ce portal \n
	*   return : l'adresse de l'objet cPortal correspondant*/
	cPortal * Declare_Portal( cSquare_Area *area , cHV_Seg& seg);

	/// ajoute au jeu les murs et les portals contenu dans un cot d'une piece
	/** 'area' : area  laquelle on va ajouter les murs et portals trouvs \n
	*   'seg'  : segment dans lequel on cherche les murs et les portals \n
	*          'seg' doit etre un cot de 'area' !!! */
	void Add_Wall_And_Portals(cSquare_Area * area , cHV_Seg& seg);

	/// ajoute au jeu les portals qui decoulent de l'ajout d'un mur au bord de cette piece
	/** 'area' : piece etudie \n
	* 'seg'  : cote de la piece etudi \n
	* 'wall' : segment de mur contenu dans 'seg' \n
	*    on va generer les portals qui sont des deux cots de 'wall' s'ils existent ! */
	void Creates_And_Add_Portals(cSquare_Area * area , cHV_Seg& seg ,cHV_Seg& wall);	

	/// divise le portail 'p' selon le segment 'seg'
	/**
	* precondition : 'seg' est contenu dans le portail \n
	* il ya les cas suivants : \n
	* suivant la position de 'seg' dans 'p', il faut modifier 'p', crer 1 ou 2 nouveaux portails \n
	*/
	void Split_Portal(cPortal * p , cHV_Seg& seg);

	/// cree un noeud entre deux portails paralelles d'un mur d'une meme piece
	/**
	* le noeud est plac a distance egale des deux portails, dcal un peu dans la piece \n
	* cela permet a une unit d'aller d'un portail a l'aute en evitant le mur  
	*/
	void Manage_Paralell_Portals(void);
};



#endif