/***************************************************************************************************/
/***************************************************************************************************/
//! \file game_graph.h
//! \brief Declare les classes utiles pour crer un graphe de noeud pour effectuer des A* search
//!    
//!	Auteur : Adrien Auclair
//!
/***************************************************************************************************/
/***************************************************************************************************/
#ifndef GAME_GRAPH_H
#define GAME_GRAPH_H

/*------------------------------------  Prototypes  ---------------------------------------------*/
class cPortal;
class cA_Star_Point;
class cParalell_Portal;

/*------------------------------------   Includes   ---------------------------------------------*/
#include "a_star.h"
#include "game_bsp.h"


/*-------------------------------------   Classes   ---------------------------------------------*/


/*************************************************************************************************/
//! noeud pour l'algorithme A* qui represente un point
/*! les noeuds usuels du A* sont des portals entre une zone et une autre ( use cA_Star_Node ) \n
    mais les points de depart et d'arrive sont des noeuds qui ne sont pas des portals (use cA_Star_Point)
/*************************************************************************************************/
class cA_Star_Point : public cA_Star_Node
{
public:
	/// constructeur
	/** 'area' : pointeur vers la cSquare_Area dans laquelle se trouve le point \n
	* 'p'    : point que ce noeud va representer */
	cA_Star_Point( cSquare_Area * area , cPoint2D& p);

	/// cost reel de ce noeud au noeud 'destination' ( en ligne droite )
	float Get_Cost( cA_Star_Node * destination);

	/// cost estim de ce noeud au noeud 'destination'
	float Get_Estimated_Cost( cA_Star_Node * destination);

	/// met ds 'voisinage' la liste des noeuds voisins 
	void  Get_List_Neighbors( cList_A_Star_Nodes & voisinage );

private:
	cSquare_Area * Area;   ///< cSquare_Area ds laquelle se trouve le point que represente ce noeud

};



/*************************************************************************************************/
//! noeud pour l'algorithme A*
/*! un portal est une porte entre deux cSquare_Area \n
    un portal s'etend sur un segment, sa position exacte est fix au milieu de ce segment
/*************************************************************************************************/
class cPortal : public cA_Star_Node
{
	friend std::ostream & operator<<(std::ostream &os, const cPortal &ex); 
public:
	/// constructeur
	/** 'id'   : id que portera ce cPortal \n
	* ce constructeur ne specifie pas entre quels cSquare_Area se trouve le portal \n
	* il faudra le specifier plus tard manuellement */
	cPortal(int id);	

	/// constructeur
	/** 'seg'  : emplacement du segment porteur de ce portal \n
	* 'area' : une cSquare_Area qui a pour porte ce portal \n
	* 'id'   : id  attribuer  ce portal */
	cPortal(cHV_Seg & seg , cSquare_Area * area , int id);	

	/// signale  ce portal qu'il est  la frontiere de la cSquare_Area 'area'
	void Add_Area( cSquare_Area * new_area); 

	/// enleve le noeud 'node' de la liste des voisins
	/** ne fait rien si 'node' ne fait pas partie de la liste des voisins */
	void Remove_Voisin( cA_Star_Node * node);
	
	/// retourne l'id de ce portal	
	inline int Get_Id() const { return Id; }; 		        
	/// cout pour aller a 'destination' qui est un noeud voisin	
	float Get_Cost( cA_Star_Node * destination);            
	/// cout estim de ce noeud au noeud 'destination'	
	float Get_Estimated_Cost( cA_Star_Node * destination);  
	/// met ds voisinage la liste des noeuds voisins		
	void  Get_List_Neighbors( cList_A_Star_Nodes & voisinage );    
	/// true si le 'portal' a les memes area1 et area2 que celui-ci	
	bool  Is_Similar( cPortal & portal);                    
	/// ajoute 'node'  la liste des noeuds voisins	
	void  Add_Voisin( cA_Star_Node * node);                 

	cSquare_Area * Area1;                                   ///< premier cSquare_Area que limite ce portal
	cSquare_Area * Area2;                                   ///< second  cSquare_Area que limite ce portal	
	cHV_Seg Segment;                                        ///< emplacement de ce portal
	cList_A_Star_Nodes Voisinage;                           ///< liste des noeuds voisins dans le graph pour le A*

protected:	
	int Id;	                                                ///< id de ce portal
	
	
};





/*************************************************************************************************/
//! portal fictif entre deux portals paralelles
/*!
        ****A*****-------W------******B******                   
                     |                                          
                     |                                          
              p1     |  p2                                      
                     |                                          
																\n
  ---W--- est un wall au milieu d'un piece \n
 alors le code genere le chemin de A vers  B pour passer de p1  p2 \n
 le probleme est que le chemin A->B est confondu avec le mur W \n
 dans ce cas, on crera un 'cParalell_Portal', il s'agira d'un noeud fictif entre  A et B \n
 qui fera que le chemin A->B passera par ce noeud et ne sera plus confondu avec le mur W \n
**************************************************************************************************/
class cParalell_Portal : public cPortal
{
public:
	/// constructeur
	/** 'seg'  : emplacement de ce portal, le portal aura comme position le milieu de ce segment\n
	*  'id'   : id de ce portal \n
	*  'area' : cSquare_Area dans lequel se trouve ce noeud */
	cParalell_Portal( cPoint2D & pos , int id,cSquare_Area * area):cPortal(id)
	{		
		//seg.Get_Middle(Position);
		//Segment = seg		
		Segment.Set(pos,pos);
		Position = pos;
		Area1=area;
		Area2=area;
	}	
};



#endif