/***************************************************************************************************/
/***************************************************************************************************/
//! \file game_manager.h
//! \brief Dfini la class cGame_Manager
//!    
//!	Auteur : Adrien Auclair
//!
//! 
/***************************************************************************************************/
/***************************************************************************************************/
#ifndef GAME_MANAGER_H
#define GAME_MANAGER_H

/*------------------------------------  Prototypes  ---------------------------------------------*/
class cGame_Manager;

/// instanciation de l'objet de gestion du jeu
extern cGame_Manager Game_Manager;

/*------------------------------------   Includes   ---------------------------------------------*/
#include <list>


#include "player.h"
#include "ennemies.h"
#include "bonus.h"
#include "game_bsp.h"
#include "gen_rnd.h"
#include "bsp_debug_X.h"


/*------------------------------------    Define    ---------------------------------------------*/
#define N_LEVELS             3
#define INITIAL_NB_LIVES     3

#define NB_MAX_BOMBS         8
#define NB_MAX_TELEPORTEURS  8
#define NB_MAX_HOLOGRAMMES   5

/// dfini les etats possibles du Game_Manager
/**
INITIALIZING   : etat du jeu initial, mis par l'appel au constructeur \n
RUNNING        : un niveau est en cours \n
READY_TO_START : un niveau est pret a etre jou, on attends que le joueur le lance ( press space) \n
LEVEL_FINISHED : le joueur vient de terminer un niveau \n
VICTORY        : le joueur vient de terminer le derier niveau \n
*/
typedef enum { INITIALIZING ,  RUNNING, GAME_OVER , READY_TO_START , LEVEL_FINISHED , VICTORY } Game_State_Type;



/*-------------------------------------   Classes   ---------------------------------------------*/

/*************************************************************************************************/
//! classe cGame_Manger est la classe de haut niveau permettant de manipuler le jeu
/*! Cette classe  different tat, voir la donne membre : Game_State \n
//! la mthode "main_loop()" switch en fonction de cette tat pour executer la bonne action \n
**************************************************************************************************/
class cGame_Manager
{
	friend class cPlayer;
	friend class cLevel_Loader;
	friend class cEnnemy;
	friend class cRed_Ball;
	friend class cBlue_Ball;
	friend class cGreen_Ball;
	friend class cUnit;
	
public :
	/// contructeur
	cGame_Manager();

	/// destructeur
	~cGame_Manager();

	/// signale au jeu qu'il doit essayer de demarer ( ne le fait qu'en mode READY_TO_START)
	void Try_Start();

	/// boucle principal du manager de jeu, agit en fonction de l'etat du manager
	void Main_Loop();	

	/*-----------------------------------------------------------------------------*/
	/// descripteur de niveau, contient le BSP et le graphe pour le A*
	cGame_World * Game_World; 
	
	Game_State_Type Game_State;                                ///< l'tat du jeu		

	bool bDisplay_Square_Areas  ;                              ///< indique si on doit afficher l'info de debug des pieces
	bool bDisplay_Ennemies_Path ;                              ///< indique si on doit afficher le chemin des ennemis
	bool bDisplay_Path_Graph    ;                              ///< indique si on doit afficher le reseua de noeuds du A*
	bool bGame_Paused           ;                              ///< indique si le jeu est mis en pause
	bool bInvincible            ;                              ///< indique si l'on est en mode invincible		


	/// les classes d'affichage
	cSprite_Display_Def * pDisplay_White_Ball ;
	cSprite_Display_Def * pDisplay_Red_Ball    ;
	cSprite_Display_Def * pDisplay_Blue_Ball   ;
	cSprite_Display_Def * pDisplay_Green_Ball  ;
	cSprite_Display_Def * pDisplay_Bomb        ; 
	cSprite_Display_Def * pDisplay_Teleporteur ;
	cSprite_Display_Def * pDisplay_Hologramme  ;
	cSprite_Display_Def * pDisplay_Cadeau      ;
	cSprite_Display_Def * pDisplay_Red_Trap    ;
	cSprite_Display_Def * pDisplay_Blue_Trap   ;
	cSprite_Display_Def * pDisplay_Green_Trap  ;
	cSprite_Display_Def * pDisplay_Coeur  ;	

	cBitmap_Display_Def * pDisplay_Game_Over;
	cBitmap_Display_Def * pDisplay_Victory ;
	cBitmap_Display_Def * pDisplay_Press_Space ;



private: 	
	std::list<cEnnemy*>              Ennemies_List;              ///< la liste des ennemis
	std::list<cInteractive_Item*>    Bonus_List;	             ///< la liste des bonus sur le jeu	
	std::vector <cEnnemy*>           Ennemies_Garbage_Collector; ///< garbage collector des ennemis	
	std::vector <cInteractive_Item*> Bonus_Garbage_Collector;    ///< garbage collector des ennemis
	
	int           Id_Level;                                    ///< numero du niveau en cours ( id du premier niveau = 0)
	float         Time_Of_Last_Bonus;                          ///< Timestamp du dernier bonus ajout au jeu par le joueur	
	int           Nb_Seconds_For_Level;                        ///< Temps donn pour finir le niveau courant
	float         Time_Last_Generated_Cadeau;                  ///< TimeStamp auquel le dernier cadeau a ete gnr	
	unsigned int  Score;                                       ///< le score du joueur
	bool          bTeleporteur_In_Game;                        ///< vrai s'il y a un teleporteur sur le jeu	
	
	cRnd_Number_Generator         Rnd_Gen ;                    ///< generateur de chiffres aletaoires	
	cGame_World_Debug_Displayer * World_Debug_Displayer;       ///< objet utilis pour afficher des infos de debug graphiques
	cPlayer * Player;                                          ///< pointeur vers le joueur
	cLevel_Loader * Level_Loader;                              ///< classe pour lire les niveaux a partir d'un fichier 

	/// temps depuis le debut du jeu, date prise lors de l'appel  update game !
	float Time_Since_Level_Start;

	/// fait evoluer le jeu
	/** cad : \n
	*   1) les ennemies mettent  jour leur chemin a suivre ( nouveau chemin ou garder l'ancien )	 \n
	*   2) les ennemies suivent leur chemin en cours et interagissent sur les bonus                  \n
	*   3) fait apparaitre des bonus si besoin est                                                   \n
	*   4) verifie que le temps n'est pas coul                                                     \n
	*/	
	void Update_Game();
	void Update_Ennemies();
	void Test_Collisions();

	/// delete les ennemies du garbage collector
	void Erase_Ennemies_Garbage_Collector(  );

	/// delete les bonus du garbage collector
	void Erase_Bonus_Garbage_Collector(  );

	/// ajoute des cadeaux au hasard ds le jeu
	void Generate_Cadeaux();

	/// affiche tous les elts du jeu	
	void Display();	
	void Display_Score();	
	void Display_Time();
	void Display_Debug_X();
	void Display_Inventaire_Player();

	/// lit le clavier et execute les actions correspondantes
	HRESULT Read_Keyboard_And_React();

	/// efface les donnes du niveau courant, pret  lire un niveau
	void Clear();

	/// remet  leur etat initial les compteurs,  appeller juste avant de lancer un niveaus
	void Reset();

	/// lit le niveau 'i' dans le jeu
	bool Load_Level( int i);

	/// start game
	/** charge le premier niveau \n
	*   le cGame_Manager est ainsi  se laisser jouer ( Read_Keyboard_And_React +  Update ) */
	void Start();	

	/// ajoute le 'new_ennemy' au jeu en cours
	/**
	  retourne true si le bonus a pu etre ajout  \n
	  false si un elt quelconque a empech d'ajouter ce bonus  sa position
	*/
	bool Add_Ennemy( Ennemy_Type bonus, int x, int y );

	/// ajoute 'ennemy' au garbage collector d'ennemis
	void Kill_Ennemy( cEnnemy * ennemy );	

	/// ajoute un bonus de type 'bonus' en x,y au jeu en cours
	/**
	  retourne true si le bonus a pu etre ajout  \n
	  false si un elt quelconque a empech d'ajouter ce bonus  sa position
	*/
	bool Add_Bonus( Interactive_Item_Type bonus, int x, int y);	

	/// ajoute 'bonus' au garbage collector de bonus
	void Remove_Bonus( cInteractive_Item * bonus);

	/// retourne vrai si la boule de centre x,y, rayon r ne touche rien (pas de murs, joueurs,enn,cadox...)
	bool Is_Free_Place( int x, int  y , int  rayon);	

	/// supprime un coeur au joueur, et agit en consequences s'il reste des coeurs ou non
	void Try_Kill_Player();

	/// charge les cSpriteDisplay def et cBitmapDisplayDefs
	void Initiate_Unit_Defs();
};



#endif